export default SASLAnonymous;
export type Connection = import("./connection.js").default;
declare class SASLAnonymous extends SASLMechanism {
}
import SASLMechanism from './sasl.js';
//# sourceMappingURL=sasl-anon.d.ts.map